/*!
 * Copyright (c) 2019 digitalSTROM AG, Switzerland
 *
 * \author Danilo Zanatta, danilo.zanatta@digitalstrom.com
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <iomanip>
#include <string.h> /* basename() */
#include <getopt.h>
#include <iterator>

#include "configupdate.h"
#include "dsf.h"
#include "CSVRead.h"

using std::cout;
using std::cerr;
using std::endl;
using std::string;

const std::string DSFConfigUpdate::DEFAULT_CONNECTION = "tcp://localhost:8442";

const int DSF_CONFIG_RETRIES = 3;

DSFConfigUpdate::DSFConfigUpdate()
    :m_bScanMode(false)
    ,m_bAllMode(false)
    ,m_bDebug(false)
    ,m_configFilename("")
{}


/*!
 * \brief print application usage
 */
void DSFConfigUpdate::usage()
{
    cout << "Usage:" << endl;
    cout << "> List:       " << basename(m_appName) << " -l -c <connection>" << endl;
    cout << "> Update:     " << basename(m_appName) << "    -c <connection> -d <dsm dsuid>" << endl;
    cout << "> Update all: " << basename(m_appName) << " -a -c <connection>" << endl;
    cout << endl;
    cout << "  Options:" << endl;
    cout << "    -f <filename> read config from csv-file" << endl;
    cout << "    -D            print debug output" << endl;
}

/*!
 * \brief parse and check parameters for consistency.
 */
bool DSFConfigUpdate::parseArguments(int argc, char** argv)
{
    int opt;
    while ((opt = getopt(argc, argv, "c:Dd:alhf:")) != -1) {
        switch (opt) {
        case 'c':
            m_connectionString = optarg;
            break;
        case 'D':
            m_bDebug = true;
            break;
        case 'd':
            m_dsfDSUID = optarg;
            {
                dsuid_t dsuid;
                if (!DSF::parseDSUIDFromString(dsuid, m_dsfDSUID)) {
                    cout << "Error: invalid dsuid: " << m_dsfDSUID << endl;
                    return false;
                }
            }
            break;
        case 'a':
            m_bAllMode = true;
            break;
        case 'l':
            m_bScanMode = true;
            break;
        case 'f':
            m_configFilename = optarg;
            break;
        case 'h':
            return false;
        default: /* '?' */
            return false;
        }
    }

    if (m_dsfDSUID.empty() && !m_bAllMode && !m_bScanMode) {
        return false;
    }
    return true;
}


/*!
 * \brief entry point for the update
 * \param argc application argument count
 * \param argv application argument vector
 * \return RETCODES::SUCCESS on success, error code otherwise
 *
 * \see RETCODES::ReturnValue
 */
int DSFConfigUpdate::run(int argc, char** argv)
{
    m_appName = argv[0];
    
#ifdef HAVE_CONFIG_H
    cout << ">> DSF Config update v" << VERSION << endl;
#endif
  
    if (!parseArguments(argc, argv)) {
        usage();
        return RETCODES::ERR_ARG;
    }

    if (m_connectionString.empty()) {
        m_connectionString = DEFAULT_CONNECTION;
        if (m_bDebug) {
            cout << "No connnection specified, defaulting to " << m_connectionString << endl;
        }
    }

    DSF dsf;
    if (!dsf.open(m_connectionString)){
        cout << "Failed to initialize dsm-api on " << m_connectionString << endl;
        return RETCODES::ERR_DSMAPI_LIB_INIT;
    }

    std::list<DSF::Device> dsfs;

    if (m_bScanMode || m_bAllMode) {
        dsf.listDevices(dsfs);
    }
    if (m_bScanMode) {
        printDeviceList(dsfs);
        if (!m_bAllMode) {
            return RETCODES::SUCCESS;
        }
    }
    if (!m_bAllMode && !m_dsfDSUID.empty()) {
        dsuid_t dsuid;
        DSF::parseDSUIDFromString(dsuid, m_dsfDSUID); ///< no need to check validity -> was already done when checking the cmdline parameters
        DSF::Device user_dsm;
        dsf.getDeviceInfo(dsuid, user_dsm);
        dsfs.push_back(user_dsm);
    }
    if (m_configFilename.length() == 0) {
        // No config file
        cerr << "I need the config filename to work. Use the option '-f <filename>'" << endl;
        return RETCODES::ERR_ARG;
    }
    std::vector<uint8_t> vNewConfig;
    CSVRead configFile(m_configFilename);
    if (!configFile.readU8(vNewConfig)) {
        cerr << "Failed to read config file " << m_configFilename << endl;
        return RETCODES::ERR_FILE_OPEN;
    }

    std::list<DSF::Device>::const_iterator it;
    for (it = dsfs.begin(); it != dsfs.end(); ++it) {

        std::string dsuidstr;
        DSF::prettyPrintDSUID(it->dsuid,dsuidstr);
        cout << "Updating config of DSF " << dsuidstr << endl;

        int trial = 1;
        int ret = -1;
        while ((ret != 0) && (trial <= DSF_CONFIG_RETRIES)) {
            ret = this->updateDSPConfig(&dsf, *it, vNewConfig);
            trial++;
        }

    }

    return RETCODES::SUCCESS;
}

int DSFConfigUpdate::updateDSPConfig(DSF *dsf, const DSF::Device dev, const std::vector<uint8_t> vConfig)
{
    uint32_t dspSwVersion;
    uint32_t dummy32;
    uint16_t dummy16;
    std::string name;
    if (!dsf->getVersion(dev.dsuid, &dummy32, &dummy32, &dspSwVersion, &dummy32, &dummy16, name)) {
        cerr << "Failed to retrieve DSP version" << endl;
        return RETCODES::ERR_DSF_VERSION_RETRIEVAL;
    }
    cout << "  Version DSP: 0x" << std::setfill('0') << std::hex << std::setw(8) << dspSwVersion << std::dec << "... ";

    if (!dsf->writeDSPConfig(dev, DSF::PHASE_ALL, vConfig)) {
        cerr << "error while writing DSP Config" << endl;
        return RETCODES::ERR_DSF_CONFIG_WRITE;
    }
    cout << "successful" << endl;

    cout << "  Verifying DSP config ... ";
    std::vector<uint8_t> vCurrentConfig;
    dsf->readDSPConfig(dev, DSF::PHASE_1, vCurrentConfig);

    if (m_bDebug) {
        cout << "[";
        if (vCurrentConfig.empty()) {
            cout << "empty";
        } else {
            for (std::vector<uint8_t>::const_iterator byte = vCurrentConfig.begin(); byte != vCurrentConfig.end(); ++byte) {
                cout << static_cast<unsigned>(*byte);
                if (byte < vCurrentConfig.end() - 1) {
                    cout << ", ";
                }
            }
        }
        cout << "] ... ";
    }

    if (vCurrentConfig != vConfig) {
        cerr << "read config does not match the written one" << endl;
        return RETCODES::ERR_DSF_CONFIG_VERIFY;
    }
    cout << "successful" << endl;

    cout << "  Rebooting DSPs ... ";
    if (!dsf->restartAllDSPs(dev)) {
        cerr << "failed" << endl;
        return RETCODES::ERR_DSF_DSP_RESTART_ERROR;
    }
    cout << "successful" << endl;

    return RETCODES::SUCCESS;
}


/*!
 * \brief write the device descriptions to the terminal
 * \param devices a list of description strings
 * \return RETCODES::SUCCESS on success, error code otherwise
 *
 * \see RETCODES::ReturnValue
 */
int DSFConfigUpdate::printDeviceList(const std::list<DSF::Device> &devices)
{
    cout << "Available devices:" << endl;
    std::list<DSF::Device>::const_iterator it;
    for (it = devices.begin(); it != devices.end(); ++it) {
        cout << "> " << it->description << endl;
    }
    return RETCODES::SUCCESS;
}

