/*!
 * Copyright (c) 2019 digitalSTROM AG, Switzerland
 *
 * \author Danilo Zanatta, danilo.zanatta@digitalstrom.com
 */

#include <iomanip>

#include <cstdlib>
#include <cstdio>
#include <cassert>
#include <sstream>
#include <string.h>

#include <digitalSTROM/dsuid.h>
#include <digitalSTROM/dsm-api-v2/dsm-api-const.h>

#include "dsf.h"


DSF::DSF() : m_opened(false),
             m_dsmApiHandle(NULL)
{
}


DSF::~DSF()
{
    close();
}


void DSF::init()
{
    m_opened = false;
    m_dsmApiHandle = NULL;
}

/*!
 * \brief setup ds485 connection to dSM
 * \param serialDevice the serial port device representing the RS485 bus
 * \return true on success
 */
bool DSF::open(const std::string& connection)
{

    m_dsmApiHandle = DsmApiInitialize();
    if (!m_dsmApiHandle) {
        return false;
    }

    int result = DsmApiOpen(m_dsmApiHandle, connection.c_str(), 0);
    if (result < 0) {
        return false;
    }

    m_opened = true;
    return true;
}


/*!
 * \brief close the ds485 connection
 */
void DSF::close()
{
    if (m_opened) {
        DsmApiCleanup(m_dsmApiHandle);
        DsmApiClose(m_dsmApiHandle);

        init();
    }
}


/*!
 * \brief connect to the dSF and get the device info
 * \return true on success
 */

bool DSF::getDeviceInfo(dsuid_t dsuid, DSF::Device &device)
{
    if (!m_opened) {
        return false;
    }

    uint32_t hwVersion;
    uint32_t armSwVersion;
    uint32_t dspSwVersion;
    uint32_t barmSwVersion;
    uint16_t apiVersion;
    std::string name;

    getVersion(dsuid, &hwVersion, &armSwVersion, &dspSwVersion, &barmSwVersion, &apiVersion, name);

    std::string dsuidString;
    prettyPrintDSUID(dsuid,dsuidString);
    std::ostringstream str;
    str << dsuidString;
    str << std::setfill('0')
        << std::hex
        << " (hw:0x" << std::setw(8) << hwVersion
        << ", arm:0x" << std::setw(8) << armSwVersion
        << ", dsp:0x" << std::setw(8) << dspSwVersion
        << ", barm:0x" << std::setw(8) << barmSwVersion
        << ", api:0x" << std::setw(4) << apiVersion << ", '"
        << name << "')";

    device.dsuid = dsuid;
    device.description = str.str();

    return true;
}

bool DSF::listDevices(std::list<Device>& devices)
{
    if (!m_opened) {
        return false;
    }

    dsuid_t device_list[63];
    int ret = DsmApiGetBusMembersPerType(m_dsmApiHandle, device_list, 63, BUS_MEMBER_GET_DSF);
    if (ret < 0) {
        return false;
    }

    for (int i = 0; i < ret; ++i) {
        DSF::Device dsf;
        getDeviceInfo(device_list[i], dsf);
        devices.push_back(dsf);
    }
    return true;
}

bool DSF::getVersion(const dsuid_t& dsuid,
                     uint32_t* hwVersion,
                     uint32_t* armSwVersion,
                     uint32_t* dspSwVersion,
                     uint32_t* barmSwVersion,
                     uint16_t* apiVersion,
                     std::string& name)
{
    if (!m_opened) {
        return false;
    }
    uint8_t nameBuf[20];
    name = "";

    int ret = dSFInfo(m_dsmApiHandle, dsuid,
                      hwVersion, armSwVersion, dspSwVersion, barmSwVersion, apiVersion,
                      NULL, nameBuf);

    if (ret != 0) {
        printf("[Error] dSFInfo returns: %d\n", ret);
        return false;
    }

    for (unsigned int i = 0; i < sizeof(nameBuf); ++i) {
        if (nameBuf[i] == 0) {
            break;
        }
        name += (char)nameBuf[i];
    }

    return true;
}

bool DSF::readDSPConfig(const Device &device, const Phase phase, std::vector<uint8_t> &vConfig) {
    if (!m_opened) {
        return false;
    }
    if (phase == PHASE_ALL) {
        // only one phase at a time
        return false;
    }

    uint8_t buffer[96];
    uint8_t size;
    int ret = dSFConfig_get_DSP_config(m_dsmApiHandle, device.dsuid, phase, &size, buffer);

    if (ret == DSF_ERROR_NO_CONFIG) { // Config empty
        vConfig.clear();
        return true;
    }

    if (ret != 0) {
        printf("[Error] dSFConfig_get_DSP_config returns %d for phase %d\n", ret, phase);
        return false;
    }

    vConfig.clear();
    for (int i=0; i<size; ++i) {
        vConfig.push_back(buffer[i]);
    }
    return true;
}

bool DSF::writeDSPConfig(const Device &device, const Phase phase, const std::vector<uint8_t> &vConfig)
{
    if (!m_opened) {
        return false;
    }
    uint8_t conf_size = static_cast<uint8_t>(vConfig.size());
    if (conf_size > 96){
        return false;
    }

    uint8_t buffer[96];
    memcpy(buffer,&(vConfig[0]),conf_size);

    std::vector<uint8_t> vPhases;
    if (phase == PHASE_ALL) {
        vPhases.push_back(PHASE_1);
        vPhases.push_back(PHASE_2);
        vPhases.push_back(PHASE_3);
    } else {
        vPhases.push_back(phase);
    }

    for (std::vector<uint8_t>::const_iterator ph = vPhases.begin(); ph != vPhases.end(); ++ph) {
        int ret = dSFConfig_set_DSP_config(m_dsmApiHandle, device.dsuid, *ph, conf_size, buffer);

        if (ret != 0) {
            printf("[Error] dSFConfig_set_DSP_config returns %d for phase %d\n", ret, *ph);
            return false;
        }
    }
    return true;
}

bool DSF::restartAllDSPs(const Device &device)
{
    if (!m_opened) {
        return false;
    }
    int ret = dSFDebugMode_restart_dsp(m_dsmApiHandle, device.dsuid, 3);
    if (ret != 0) {
        printf("[Error] dSFDebugMode_restart_dsp returns: %d\n", ret);
        return false;
    }
    return true;
}

void DSF::prettyPrintDSUID(const dsuid_t& dsuid, std::string& dsuidString)
{
    std::ostringstream str;
    for (int j = 0; j < DSUID_SIZE; ++j) {
        str << (j ? ":" : "")
            << std::setw(2)
            << std::setfill('0')
            << std::hex
            << (int)dsuid.id[j];
    }
    dsuidString = str.str();
}

bool DSF::parseDSUIDFromString(dsuid_t& dsuid, const std::string& dsuidString)
{
    if (dsuidString.size() != (DSUID_SIZE-1)*3+2) {
        return false;
    }

    for (unsigned long i = 0; i < DSUID_SIZE; ++i) {
        std::string sub = dsuidString.substr(i*3, i*3+2);
        dsuid.id[i] = static_cast<unsigned char>(strtol(sub.c_str(), NULL, 16));
    }

    return true;
}
